#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

HTMLWriter.py

Copyright 2023 by Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.

"""
import os
import zipfile

from libgutenberg.Logger import debug, exception, info, error, warning
import libgutenberg.GutenbergGlobals as gg

from ebookmaker.writers import HTMLWriter as BaseHTMLWriter
from ..EbookConverter import make_output_filename, FILENAMES

BaseHTMLWriter.FILENAMES = FILENAMES

def arcurl(job, outfile):
    if outfile.startswith(job.outputdir):
        return outfile[len(job.outputdir) + 1:]
    else:
        info(job.outputdir, outfile)
        return outfile

class Writer(BaseHTMLWriter.Writer):
    """ Class for writing HTML files, including Zip bundle. """


    def build(self, job):
        """ Build HTML file. """
        old_credit = job.dc.credit
        info(f"credit was db: {job.dc.credit}")
        super().build(job)
        try:
            # now zip up the files
            zipfilename = os.path.join(job.outputdir, make_output_filename('zip', job.ebook))
            with zipfile.ZipFile(zipfilename, 'w', zipfile.ZIP_DEFLATED) as outzipfile:
                for p in job.spider.parsers:
                    outfile = gg.normalize_path(self.outputfileurl(job, p.attribs.url))
                    if outfile.endswith('/'):
                        # main file, special handling
                        outfile = os.path.join(outfile, job.outputfile)

                    try:
                        os.stat(outfile)
                        dummy_name, ext = os.path.splitext(outfile)
                        info(' Adding file: %s as %s' % (outfile, arcurl(job, outfile)))
                        outzipfile.write(outfile, arcurl(job, outfile),
                                    zipfile.ZIP_STORED if ext in ['.zip', '.png', '.jpeg', '.jpg']
                                    else zipfile.ZIP_DEFLATED)
                    except OSError:
                        warning ('build zip: Cannot add file %s', outfile)
        except Exception as what:
            exception("Error making zip %s: %s" % (outzipfile, what))
            if os.access(outfile, os.W_OK):
                os.remove(outfile)
            raise what
        
        info("Done making zip: %s" % job.outputfile)
        
        if job.dc.credit and job.dc.credit != old_credit:
            job.dc.add_attribute(job.dc.book, job.dc.credit, marc=508)
            job.dc.session.commit()
            info(f"set credit to db: {job.dc.credit}")
        



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
RDFWriter.py

Copyright 2009,2014 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Builds RDF entry for ebook.

"""

import os.path

import rdflib
from rdflib import BNode, URIRef, Literal

import libgutenberg.GutenbergGlobals as gg
from libgutenberg.GutenbergGlobals import NS, NSURI, Struct
from libgutenberg.Logger import info
from ebookmaker import writers

RNS = Struct ()

for (_prefix, _uri) in gg.NSMAP.items ():
    setattr (RNS, _prefix, rdflib.namespace.Namespace (_uri))

EXPORTED_NAMESPACES = 'cc dcam dcterms marcrel pgterms rdf rdfs'.split ()

rdflib.term.bind (NSURI.dcterms.IMT, gg.DCIMT)


class Writer (writers.BaseWriter):
    """ A class to write an RDF Bibrec. """


    def __init__ (self):
        super (Writer, self).__init__ ()
        self.graph = rdflib.Graph ()


    def build (self, job):
        """ Build RDF record for ebook. """

        info ("Making   %s" % job.outputfile)
        fn = os.path.join (job.outputdir, job.outputfile)

        # build tree
        self.from_dc (job.dc)

        with open (fn, "wb") as fp:
            self.serialize (fp)

        info ("Done     %s" % job.outputfile)


    def serialize (self, fp):
        """ Serialize RDF graph to xml. """

        g = self.graph

        for prefix in EXPORTED_NAMESPACES:
            g.namespace_manager.bind (prefix, gg.NSMAP[prefix])

        g.serialize (fp, format = "pretty-xml",
                     base = str (NS.pg), xml_base = str (NS.pg))


    def append (self, s, p, o):
        """ Helper. """

        self.graph.add ( (s, p, o) )


    def vocabulary (self, parent, rows, term_name, vocabulary_encoding_scheme):
        """ Add term from vocabulary to node. """

        for row in rows:
            resource = BNode () # rdf:Description
            self.append (resource, RNS.dcam.memberOf, vocabulary_encoding_scheme)
            self.append (resource, RNS.rdf.value, Literal (row))
            self.append (parent, term_name, resource)


    def datatype (self, parent, rows, term_name, datatype):
        """ Add term with datatype to node. """

        for row in rows:
            resource = BNode () # rdf:Description
            self.append (resource, RNS.rdf.value, Literal (row, datatype=datatype))
            self.append (parent, term_name, resource)


    def from_dc (self, dc):
        """ Create RDF graph from DublinCore struct. """

	# gbn 2017-01-29: Updated license
        # lic = URIRef ('http://www.gnu.org/licenses/gpl.html')
        lic = URIRef ('https://creativecommons.org/publicdomain/zero/1.0/')
        this_doc = URIRef ('')

        self.append (this_doc, RNS.cc.license, lic)
        self.append (this_doc, RNS.rdf.type,   RNS.cc.Work)
        self.append (this_doc, RNS.rdfs.comment, Literal (
            '''Archives containing the RDF files for *all* our books can be downloaded at
            https://www.gutenberg.org/wiki/Gutenberg:Feeds#The_Complete_Project_Gutenberg_Catalog'''
        ))


        # book node

        book = Literal (RNS.ebook[str (dc.project_gutenberg_id)])
        self.append (book, RNS.rdf.type, RNS.pgterms.ebook)


        # publisher, copyright and release date

        self.append (book, RNS.dcterms.publisher, Literal ('Project Gutenberg'))
        self.append (book, RNS.dcterms.license,   RNS.pg.license)
        self.append (book, RNS.dcterms.issued,    Literal (dc.release_date))
        self.append (book, RNS.dcterms.rights,    Literal (dc.rights))
        self.append (book, RNS.pgterms.downloads, Literal (dc.downloads))


        # authors
        # for a list of relator codes see:
        # http://www.loc.gov/loc.terms/relators/

        for author in dc.authors:
            # make author node
            a = RNS.pgagents[str (author.id)]
            self.append (a, RNS.rdf.type, RNS.pgterms.agent)
            self.append (a, RNS.pgterms.name, Literal (author.name))
            if author.birthdate:
                self.append (a, RNS.pgterms.birthdate, Literal (author.birthdate))
            if author.deathdate:
                self.append (a, RNS.pgterms.deathdate, Literal (author.deathdate))

            # aliases
            for alias in author.aliases:
                self.append (a, RNS.pgterms.alias, Literal (alias.alias))

            # web pages
            for webpage in author.webpages:
                p = URIRef (webpage.url)
                self.append (p, RNS.dcterms.description, Literal (webpage.description))
                self.append (a, RNS.pgterms.webpage, p)

            # append author node to book node
            if author.marcrel == 'aut' or author.marcrel == 'cre':
                self.append (book, RNS.dcterms.creator, a)
            else:
                self.append (book, RNS.marcrel[str (author.marcrel)], a)


        # titles, notes

        marc2dc = {
            '240': RNS.dcterms.alternative,
            '245': RNS.dcterms.title,
            '246': RNS.dcterms.alternative,
            '500': RNS.dcterms.description,
            '505': RNS.dcterms.tableOfContents,
            }

        for marc in dc.marcs:
            try:
                pred = marc2dc[marc.code]
                self.append (book, pred, Literal (marc.text))

                if pred == RNS.dcterms.title:
                    dc.title = marc.text

            except KeyError:
                # other  marc code
                self.append (book, RNS.pgterms[str ('marc' + marc.code)], Literal (marc.text))


        # languages (datatype)

        self.datatype (book, [language.id for language in dc.languages],
                       RNS.dcterms.language, RNS.dcterms.RFC4646)

        # subjects (vocabulary)

        self.vocabulary (book, [subject.subject for subject in dc.subjects],
                         RNS.dcterms.subject, RNS.dcterms.LCSH)

        # LoCC (vocabulary)

        self.vocabulary (book, [locc.id for locc in dc.loccs],
                         RNS.dcterms.subject, RNS.dcterms.LCC)

        # types (vocabulary)
        self.vocabulary (book, [cat.id for cat in dc.dcmitypes],
                         RNS.dcterms.type, RNS.dcterms.DCMIType)

        # bookshelves (vocabulary)

        self.vocabulary (book, [bs.bookshelf for bs in dc.bookshelves],
                         RNS.pgterms.bookshelf, RNS.pgterms.Bookshelf)

        # files

        for file_ in dc.files + dc.generated_files:
            # file node
            f = URIRef (RNS.pg[str (file_.url)])
            self.append (f, RNS.rdf.type, RNS.pgterms.file)

            # two-way link to ebook node
            self.append (book, RNS.dcterms.hasFormat, f)
            self.append (f, RNS.dcterms.isFormatOf,
                         URIRef (RNS.ebook[str (dc.project_gutenberg_id)]))

            # extent, modified
            self.append (f, RNS.dcterms.extent,   Literal (file_.extent))
            self.append (f, RNS.dcterms.modified, Literal (file_.modified))

            # mediatype (vocabulary)
            # note: avoid <built-in method format of Namespace>
            self.vocabulary (f, file_.mediatypes, RNS.dcterms['format'], RNS.dcterms.IMT)



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
MastodonWriter.py

Copyright 2023 by Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.

Toot about new ebooks.

"""

import os
import uuid

import requests

from requests import RequestException

from libgutenberg.Logger import exception
from libgutenberg.GutenbergGlobals import SkipOutputFormat
from libgutenberg import GutenbergDatabase

from ebookmaker import writers


URL = "https://mastodon.social/api/v1/statuses"

class Writer(writers.BaseWriter):
    """ Class to toot about new ebooks.

    """

    __instance = None

    def __new__(cls):
        """ Make Writer a singleton. """

        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
        return cls.__instance


    def __init__(self):

        super(Writer, self).__init__()

        self.access_token = os.environ.get('MASTODON_ACCESS_TOKEN')
        if not self.access_token:
            raise SkipOutputFormat ('Mastodon credentials not found.')


    def build(self, job):
        """ Toot. """

        if job.dc.project_gutenberg_id < 70800:
            # don't toot about old books
            return

        db = GutenbergDatabase.Database()
        db.connect()
        c = db.get_cursor()

        c.execute("select * from tweets where media = 'Mastodon' and fk_books = %(fk_books)s",
                   {'fk_books': job.dc.project_gutenberg_id})

        rows = c.fetchone()
        if rows:
            # already tooted
            return

        try:
            title = job.dc.make_pretty_title(500 - 84)  # 500 is max length
            headers = {'Authorization': 'Bearer %s' % self.access_token,
                       'Idempotency-Key': str(uuid.uuid4())}
            toot = "New #ebook @gutenberg_org@mastodon.social: %s https://www.gutenberg.org/ebooks/%d" % (
                title, job.dc.project_gutenberg_id)
            lang_id = job.dc.languages[0].id if len(job.dc.languages) else 'en'
            data = {'status': toot,
                    'visibility': 'public',
                    'language': lang_id}
            r = requests.post(URL, headers=headers, data=data)
            r.raise_for_status()

            c.execute('start transaction')
            c.execute(
                """\
                insert into tweets (fk_books, media, time)
                values (%(fk_books)s, 'Mastodon', now ())
                """, {'fk_books': job.dc.project_gutenberg_id})
            c.execute('commit')

        except (RequestException) as what:
            exception('MastodonWriter: could not toot about ebook: %s' % what)

        except (GutenbergDatabase.DatabaseError) as what:
            c.execute('rollback')
            exception('MastodonWriter: could not write to database (#%s)' % what)



#!/usr/bin/env python3
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

ReloadWorkflow.py

arguments: a list of ebook numbers

Copyright 2022 by Eric Hellman and Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.

looks for the newest workflow json file, and reloads the data into the database, overwriting whatever is there. Use with caution!!

"""
import os
import sys

from libgutenberg import Logger
from libgutenberg.DublinCoreMapping import DublinCoreObject
from .FileInfo import get_workflow_file, WORKFLOW_LOG_DIR

def get_newest_workflow_file(ebook):
    current = get_workflow_file(ebook)
    if current:
        return current
    for rev in reversed(range(0,10)):
        archived = os.path.join(WORKFLOW_LOG_DIR, 'backup', '%s.json.%s' % (ebook, rev))
        if os.path.exists(archived):
            return archived
    archived = os.path.join(WORKFLOW_LOG_DIR, 'backup', '%s.json' % ebook)
    if os.path.exists(archived):
        return archived

def reload_workflow_file(ebook):
    workflow_file = get_newest_workflow_file(ebook)
    if not workflow_file:
        print(f'No workflow file for {ebook}')
        return
    dc = DublinCoreObject()
    dc.load_book(ebook)
    with open(workflow_file, 'r') as fp:
        data = fp.read()
    try:
        dc.load_from_pgheader(data)
    except ValueError:
        print(f'Tried {workflow_file}; was it a valid workflow file?')
        Logger.error("Tried %s; was it a valid workflow file?", workflow_file)
        return
    dc.save(updatemode=1)


def main():
    Logger.setup(Logger.LOGFORMAT, 'fileinfo.log')
    Logger.set_log_level(2)

    if len(sys.argv) > 1:
        for arg in sys.argv[1:]:
            try:
                Logger.ebook = int(arg)
                print(f'loading metadata for {arg}')
                reload_workflow_file(int(arg))
            except ValueError: # no int
                print('need a book number to load')

    else:
        print('need a book number to load')

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

CoverpageWriter.py

Copyright 2010 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Writes coverpage thumbnails.

"""


import os

from libgutenberg.Logger import info, exception

from ebookmaker import writers


MAX_IMAGE_SIZE  =  63 * 1024  # in bytes
MAX_IMAGE_DIMEN = (200, 300)  # in pixels

MAX_THUMB_SIZE  =  16 * 1024  # in bytes
MAX_THUMB_DIMEN = ( 66, 100)  # in pixels

class Writer (writers.BaseWriter):
    """ Class that writes coverpage thumbs. """

    dimensions = {
        'cover.medium': (MAX_IMAGE_SIZE, MAX_IMAGE_DIMEN),
        'cover.small':  (MAX_THUMB_SIZE, MAX_THUMB_DIMEN),
        }


    def build (self, job):
        """ Build coverpage thumbs """

        info ("Making   %s" % job.outputfile)

        try:
            for p in job.spider.parsers:
                if hasattr (p, 'resize_image') and 'icon' in p.attribs.rel:
                    dimen = self.dimensions [job.type]
                    np = p.resize_image (dimen[0], dimen[1], 'jpeg')

                    fn = os.path.join (job.outputdir, job.outputfile)
                    with open (fn, 'wb') as fp:
                        fp.write (np.serialize ())

                    info ("Found coverpage %s" % p.attribs.url)
                    # if job.ebook and hasattr (job.dc, 'register_coverpage'):
                    #     job.dc.register_coverpage (job.ebook, p.attribs.url)
                    break

            info ("Done     %s" % job.outputfile)

        except Exception as what:
            exception ("Error building coverpage: %s" % what)
            raise



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""
EbookConverter.py

Copyright 2009-2014 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Driver for the ibiblio environment.  Queries the PG database and finds
out which files to convert, then passes them to Ebookmaker.

"""

import argparse
import collections
import configparser
import datetime
import itertools
import logging
import os
import subprocess
import sys

from six.moves import urllib, cPickle
import setproctitle

from libgutenberg import DBUtils, Logger
from libgutenberg.GutenbergFiles import remove_file_from_database, store_file_in_database
from libgutenberg.GutenbergGlobals import Struct
from libgutenberg.Logger import critical, info, debug, warning, error, exception


from ebookmaker import CommonCode
from ebookmaker.CommonCode import Options

from ebookconverter import Candidates, Notifier
from ebookconverter.Version import VERSION

options = Options()

CONFIG_FILES = ['/etc/ebookconverter.conf', os.path.expanduser('~/.ebookconverter')]

NON_UTF_TXTS = ('txt/windows-125*', 'txt/iso-8859-*', 'txt/big5', 'txt/ibm*', 'txt/*')
ALL_TXTS = ('txt/utf-8', ) + NON_UTF_TXTS
ALL_HTM =  ('html/utf-8', 'html/windows-125*', 'html/iso-8859-*', 'html/*')
ALL_SRCS = ('rst/*',) + ALL_HTM + ALL_TXTS

PREFERRED_INPUT_FORMATS = {

    # current practice is not to upload rst.master yet
    'rst.gen': ('rst/*', ),

    # epub readers should be able to handle unicode,
    # prefer the big charsets but accept any charset
    'epub.images': ALL_SRCS,
    'epub3.images': ALL_SRCS,

    'kindle.images': ALL_SRCS,
    'kindle.noimages': ALL_SRCS,
    'kf8.images': ALL_SRCS,

    # html is created from rst files or text files
    'html.images': ALL_HTM + ('rst/*',) + ALL_TXTS,

    # utf txt is created from text files
    'txt.utf-8': ('rst/*',) + NON_UTF_TXTS,

    # pdf is created only from rst
    'pdf.images': ('rst/*',),

    # picsdir only if pdf or html are created
    'picsdir.images': ('rst/*',),

    # coverpage (a cover will be generated, whatever)
    'cover.medium': ('rst/*', 'html/*', 'readme/*', 'txt/*', 'tex/*',),

}

PREFERRED_INPUT_FORMATS['html.noimages']      = PREFERRED_INPUT_FORMATS['html.images']
PREFERRED_INPUT_FORMATS['epub.noimages']      = PREFERRED_INPUT_FORMATS['epub.images']
PREFERRED_INPUT_FORMATS['pdf.noimages']       = PREFERRED_INPUT_FORMATS['pdf.images']
PREFERRED_INPUT_FORMATS['picsdir.noimages']   = PREFERRED_INPUT_FORMATS['picsdir.images']
PREFERRED_INPUT_FORMATS['cover.small']        = PREFERRED_INPUT_FORMATS['cover.medium']
PREFERRED_INPUT_FORMATS['null']               = PREFERRED_INPUT_FORMATS['epub.images']

# don't build a type if we've already made one
EXCLUSIONS = {
    'epub.images': ('epub.dp',   ),
    'epub.noimages': ('epub.dp',   ),
    'html.noimages': ('html/*',    ),
}

FILENAMES = {
    'html.noimages':    'pg{id}.html',
    'html.images':      'pg{id}-images.html',
    'epub.noimages':    'pg{id}.epub',
    'epub.images':      'pg{id}-images.epub',
    'epub3.images':     'pg{id}-images-3.epub',
    'kindle.noimages':  'pg{id}.mobi',
    'kindle.images':    'pg{id}-images.mobi',
    'kf8.images':       'pg{id}-images-kf8.mobi',
    'pdf.noimages':     'pg{id}.pdf',
    'pdf.images':       'pg{id}-images.pdf',
    'txt.utf-8':        'pg{id}.txt',
    'rdf':              'pg{id}.rdf',
    'rst.gen':          'pg{id}.rst',
    'cover.small':      'pg{id}.cover.small.jpg',
    'cover.medium':     'pg{id}.cover.medium.jpg',
    'qrcode':           'pg{id}.qrcode.png',
    'zip':              'pg{id}-h.zip',
    'logfile':          'pg{id}.converter.log',      # .converter because of latex log conflicts
}
GENERIC_FILENAME = 'pg{id}.generic'

DEPENDENCIES = collections.OrderedDict((
    ('everything',      ('all', 'kindle.noimages','facebook', 'bluesky', 'mastodon', 'update')),
    ('all',             ('html', 'epub', 'kindle', 'epub3', 'kf8', 'pdf', 'txt', 'rst',
                         'cover', 'qrcode', 'rdf')),
    ('html',            ('html.images',    'html.noimages')),
    ('epub',            ('epub.images',    'epub.noimages')),
    ('epub3',           ('epub3.images', )),
    ('kindle',          ('kindle.images',)),
    ('kf8',             ('kf8.images', )),
    ('pdf',             ('pdf.images',     'pdf.noimages')),
    ('txt',             ('txt.utf-8',      'txt.iso-8859-1', 'txt.us-ascii')),
    ('rst',             ('rst.gen', )),
    ('kindle.noimages', ('epub.noimages', )),
    ('kindle.images',   ('epub.images', )),
    ('html.noimages',   ('picsdir.noimages', )),
    ('html.images',     ('picsdir.images', )),
    ('pdf.noimages',    ('picsdir.noimages', )),
    ('pdf.images',      ('picsdir.images', )),
    ('rst.gen',         ('picsdir.images', )),
    ('cover',           ('cover.medium',   'cover.small')),
))

BUILD_ORDER = """
picsdir.images picsdir.noimages
rst.gen
txt.utf-8
html.images html.noimages
epub.noimages kindle.noimages pdf.noimages
cover.small cover.medium
epub.images kindle.images pdf.images
epub3.images kf8.images
qrcode rdf
facebook bluesky mastodon
update
null
""".split()

MAX_CANDIDATE_SIZE = {'epub': 16, 'epub3': 16, 'kindle': 16, 'kf8': 16}

def make_output_filename(type_, ebook = 0):
    """ Make a suitable filename for output type. """

    return FILENAMES.get(type_, GENERIC_FILENAME).format(id = ebook)


class Maker():
    """ Helper class """

    def __init__(self, ebook):
        self.ebook = ebook


    def get_cache_dir(self):
        """ return the cache directory for this ebook """
        return os.path.join(options.config.CACHEDIR, "%d" % self.ebook)

    def get_cache_loc(self):
        """ return the cache loc for this ebook """
        return os.path.join(options.config.CACHELOC, "%d" % self.ebook)

    def should_do_job(self, job, candidate):
        """ 
        Return True if outputfile newer than candidate exists
        and the candidate exists or if no candidate needed
        """
        if len(PREFERRED_INPUT_FORMATS.get(job.type, {})) == 0:
            # doesn't need a source file
            return True
 
        if not candidate:
            # happens if the candidates are too large
            return False

        if candidate.generated:
            candidate_path = os.path.join('/', self.get_cache_dir(), candidate.archive_path)
        else:
            candidate_path = os.path.join(options.config.FILESDIR, candidate.archive_path)
                   
        if not candidate.generated and not os.path.exists(candidate_path[7:]):
            warning('expected file %s not found. job skipped.', candidate_path[7:])
            return False

        path = os.path.join(job.outputdir, job.outputfile)

        if os.path.isfile(path):
            mtime_cand = candidate.modified
            mtime_epub = datetime.datetime.fromtimestamp(os.path.getmtime(path))

            debug('mtime cand:  %s' % mtime_cand)
            debug('mtime dest:  %s' % mtime_epub)

            if mtime_cand < mtime_epub:
                if job.type in options.build:
                    info('Making   %s: user requested build.' % job.outputfile)
                    return True
                info('Skipping %s: target newer than candidate.' % job.outputfile)
                return False

            info('Making   %s: target out of date.' % job.outputfile)
            return True

        info('Making   %s: does not exist.' % job.outputfile)
        return True


    def remove_type(self, type_):
        """ Remove file for type. """

        fn = os.path.join(self.get_cache_dir(), make_output_filename(type_, self.ebook))
        try:
            os.remove(fn)
            debug("Removed file from disk: %s" % fn)
        except OSError:
            pass
        try:
            os.remove(fn + '.gz')
            debug("Removed file from disk: %s.gz" % fn)
        except OSError:
            pass

        fn = os.path.join(self.get_cache_loc(), make_output_filename(type_, self.ebook))

        if options.shadow:
            debug("If not in shadow, would have removed file from database: %s" % fn)
        else:
            remove_file_from_database(fn)
            debug("Removed file from database: %s" % fn)


    def mk_job_queue(self):
        """ Make job queue for one ebook. """

        cf = Candidates.Candidates()
        all_candidates = cf.read_from_database(self.ebook)
        job_queue = []
        f = lambda x: x.format
        debug("All Candidates: %s" % ' '.join(map(f, all_candidates)))

        for type_ in options.make:
            debug("Trying: %s ..." % type_)

            candidate_types = PREFERRED_INPUT_FORMATS.get(type_, {})
            job = CommonCode.Job(type_)
            job.ebook = self.ebook
            job.outputdir  = self.get_cache_dir()
            job.outputfile = make_output_filename(type_, self.ebook)
            job.logfile = make_output_filename('logfile', self.ebook)

            candidates = all_candidates[:]

            if type_ in EXCLUSIONS and cf.filter_sort(
                    EXCLUSIONS[type_], [c for c in candidates if not c.generated], f):
                info('%s is already posted.' % type_)
                if not options.dry_run:
                    self.remove_type(type_)
                continue

            candidate = None
            if len(candidate_types) > 0:
                if DBUtils.is_not_text(self.ebook) and job.maintype != 'cover':
                    info("Book is not a text book. Skipping %s ..." % type_)
                    continue

                candidates = cf.filter_sort(candidate_types, candidates, f)

                if not candidates:
                    info('No input file found for type: %s' % type_)
                    if not options.dry_run:
                        self.remove_type(type_) # clean leftovers
                    continue

                candidate = candidates[0]
                # oom-killer safeguard
                if candidate.extent > MAX_CANDIDATE_SIZE.get(job.maintype, 32) * 1024 * 1024:
                    warning(f'Skipping {candidate.archive_path} for {type_}: file too big' )
                    continue

                job.url = os.path.join(options.config.FILESDIR, candidate.archive_path)
                info(f'type: {type_}; job.url: {job.url}' )
                # allow any file below basedir of ebook
                job.include = [ os.path.join(
                    options.config.FILESDIR, os.path.dirname(candidate.archive_path) + '/*') ]
                job.max_depth = 3
                job.source = urllib.parse.urljoin(options.config.PGURL, candidate.archive_path)
                job.opf_identifier = (urllib.parse.urljoin(
                    options.config.BIBREC + '/', str(self.ebook)))

            if self.should_do_job(job, candidate):
                job_queue.append(job)

                new_candidate = Struct()
                new_candidate.archive_path = os.path.join(job.outputdir, job.outputfile)
                new_candidate.format = job.type + '/unknown'
                new_candidate.modified = datetime.datetime.now()
                new_candidate.extent = 0
                new_candidate.generated = True
                all_candidates.insert(0, new_candidate)

        return job_queue


def run_job_queue(job_queue):
    def add_file_to_db(filename, filetype, ebook_no):
        if os.access(filename, os.R_OK):
            if options.shadow:
                debug('if not in shadow, would have stored %s in database.', filename)
            else:
                debug('adding %s to database.', filename)
                store_file_in_database(ebook_no, filename, filetype)
            mod_timestamp = datetime.datetime.fromtimestamp(os.path.getmtime(filename))
            if datetime.date.today() - mod_timestamp.date() > datetime.timedelta(1):
                critical('Failed to build new file: %s', filename)
            for ext in ['.gz', '.gzip', '.utf8']:
                if os.access(filename + ext, os.W_OK):
                    os.remove(filename + ext)
        elif '.generic' not in filename:
            critical('Failed to build file: %s', filename)


    """ Run EbookMaker for all jobs in the queue. """

    for job in job_queue:
        try:
            os.mkdir(job.outputdir, 0o775)
        except OSError: # directory exists
            pass

    verbosity = ''
    if options.verbose:
        verbosity = '-' + 'v' * options.verbose

    try:
        ebm_params = [
            options.config.EBOOKMAKER, verbosity,
            "--extension-package", "ebookconverter.writers",
            "--validate" if options.validate else None,
            "--notify" if options.notify else None,
            "--jobs", "no_such_url",
        ]
        ebm_params = [prm for prm in ebm_params if prm]
        ebm = subprocess.Popen(
            ebm_params,
            stdin  = subprocess.PIPE,
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE
        )
    except OSError as what:
        error("%s %s" % (options.config.EBOOKMAKER, what))
        raise

    debug("Calling Ebookmaker ...")
    stdout, stderr = ebm.communicate(cPickle.dumps(job_queue))
    info("Ebookmaker returned code: %d." % ebm.returncode)
    debug(stdout.decode(sys.stdout.encoding))
    debug(stderr.decode(sys.stderr.encoding))
    if ebm.returncode == 0:
        for job in job_queue:
            if job.type == 'qrcode':
                continue
            Logger.ebook = job.ebook
            filename = os.path.join(job.outputdir, job.outputfile)
            add_file_to_db(filename, job.type, job.ebook)
            if job.type == 'html.images':
                zipfilename = os.path.join(job.outputdir, make_output_filename('zip', job.ebook))
                # also add the zip
                add_file_to_db(zipfilename, None, job.ebook)
                
    else:
        critical('returncode was %s', ebm.returncode)


def add_local_options(ap):
    """ Add local options to commandline. """

    ap.add_argument(
        '--version',
        action='version',
        version = "%%(prog)s %s" % VERSION
    )

    ap.add_argument(
        "--pidfile",
        metavar  = "FILE",
        dest     = "pidfile",
        action   = "store",
        default  = "/tmp/ebookconverter.pid",
        help     = "use pid file (default: /tmp/ebookconverter.pid)")

    ap.add_argument(
        "--make",
        metavar = "TYPES",
        dest    = "make",
        choices = CommonCode.add_dependencies(['everything'], DEPENDENCIES),
        default = [],
        action  = "append",
        help    = "types to make if source newer than target")

    ap.add_argument(
        "--build",
        metavar = "TYPES",
        dest    = "build",
        choices = CommonCode.add_dependencies(['everything'], DEPENDENCIES),
        default = [],
        action  = "append",
        help    = "types to make unconditionally")

    ap.add_argument(
        "--range",
        dest     = "range",
        default  = "",
        help     = "ebook range to convert: eg. 1,2-4,6,8-")

    ap.add_argument(
        "--goback",
        metavar = "HOURS",
        dest    = "goback",
        type    = int,
        default = 0,
        action  = "store",
        help    = "convert only ebooks younger than HOURS hours")

    ap.add_argument(
        "--top",
        metavar = "N",
        dest    = "top",
        type    = int,
        default = 0,
        action  = "store",
        help    = "convert only the N most-downloaded ebooks")

    ap.add_argument(
        "--jobs",
        metavar = "N",
        dest    = "jobs",
        type    = int,
        default = 1,
        action  = "store",
        help    = "send N ebooks per job to ebookmaker(default: 1)")

    ap.add_argument(
        "--fk-filetype",
        metavar = "TYPE",
        dest    = "fk_filetype",
        default = [],
        action  = "store",
        help    = "convert only books that have a file of this filetype eg rst or html")

    ap.add_argument(
        "-n", "--dry-run",
        dest    = "dry_run",
        action  = "store_true",
        help    = "don't actually run Ebookmaker; just print the commands.")

    ap.add_argument(
        "--shadow",
        dest    = "shadow",
        action  = "store_true",
        help    = "run, but don't change postgres.")

    ap.add_argument(
        "--stop",
        dest    = "stop_on_errors",
        action  = "store_true",
        help    = "stop immediately on errors.")


def fix_option_range(options, last_ebook):
    """ Convert a range from user input into list of ebook nos. """

    if not options.range:
        return

    res = []

    try:
        for value in options.range.split(','):
            r = value.split('-')
            try:
                if len(r) == 2:
                    r[0] = int(r[0]) if r[0] else 1
                    if r[1] == '':
                        r[1] = last_ebook
                    else:
                        r[1] = int(r[1])

                    if r[0] > r[1]:
                        raise ValueError

                    res.extend(range(r[0], r[1] + 1))
                elif len(r) == 1:
                    res.append(int(value))
                else:
                    raise ValueError
            except ValueError:
                error("error in range item %s", value)
                continue

    except ValueError:
        raise ValueError("error in range parameter")

    options.range = res


def pretty_print_list(list_of_ebook_nos):
    """ Return list of ebooks as pretty string. """

    result = []
    for dummy_k, group in itertools.groupby(
            enumerate(list_of_ebook_nos), lambda x: x[0] - x[1]):
        subrange = [str(g[1]) for g in group]
        if len(subrange) <= 2:
            result += subrange
        else:
            result.append('%s-%s' % (subrange[0], subrange[-1]))
    return ' '.join(result)


def config():
    """ Process config files and commandline params. """

    ap = argparse.ArgumentParser(prog = 'EbookConverter')
    CommonCode.add_common_options(ap, CONFIG_FILES[1])
    add_local_options(ap)
    CommonCode.set_arg_defaults(ap, CONFIG_FILES[1])

    global options
    options.update(vars(CommonCode.parse_config_and_args(
        ap,
        CONFIG_FILES[0],
        {
            'proxies' : None,
            'logfile': None,
            'pgvpncmd': None,
            'ebookmaker': 'ebookmaker',
            'timestamp': datetime.datetime.today().isoformat()[:19],
        }
    )))

def grouper(iterable, n, fillvalue = None):
    """ Itertools recipe: Collect data into fixed-length chunks or blocks """
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return itertools.zip_longest(*args, fillvalue = fillvalue)


def main():
    """ Main program. """

    try:
        config()
    except configparser.Error as what:
        error("Error in configuration file: %s", str(what))
        return 1

    Logger.base_logfile = options.config.LOGFILE
    Logger.notifier = CommonCode.queue_notifications
    Logger.setup(
        Logger.LOGFORMAT,
        loglevel=options.verbose,
    )

    if options.verbose >= 1 and options.config.LOGFILE:
        print("Logging to: %s" % options.config.LOGFILE)

    debug("Using config file: %s" % options.config_file)

    # write pidfile
    try:
        fd = os.open(options.pidfile, os.O_WRONLY | os.O_CREAT | os.O_EXCL)
        debug("Writing pid file")
        os.write(fd, ("%d\n" %  os.getpid()).encode('us-ascii'))
        os.close(fd)
    except OSError:
        info("Not running: pidfile exists.")
        sys.exit(2)

    info("Program start")

    try:
        fix_option_range(options, DBUtils.last_ebook())
    except ValueError:
        info("Not running: pidfile exists.")
        sys.exit(2)


    if options.goback:
        interval = datetime.datetime.now() - datetime.timedelta(hours=options.goback)
        pks = DBUtils.recent_books(interval)
        options.range = sorted(pks)

    if options.top:
        pks = DBUtils.top_books(options.top)
        pks = pks.intersection(options.range)
        options.range = sorted(pks)

    if options.fk_filetype:
        pks = DBUtils.filetype_books(options.fk_filetype)
        pks = pks.intersection(options.range)
        options.range = sorted(pks)

    info("Processing %d ebooks" % len(options.range))
    info("Processing ebooks: %s" % pretty_print_list(options.range))

    if options.goback:
        # make sure the books just posted get done first
        options.range.sort(reverse=True)

    options.make += options.build

    options.make  = CommonCode.add_dependencies(options.make,  DEPENDENCIES, BUILD_ORDER)
    options.build = CommonCode.add_dependencies(options.build, DEPENDENCIES, BUILD_ORDER)

    info("Making:   %s" % " ".join(options.make))
    info("Building: %s" % " ".join(options.build))

    done_books  = []

    try:
        for group in grouper(options.range, options.jobs):
            job_queue = []
            last = 0
            progress = len(done_books) * 100 // len(options.range)
            info("Progress: %d%% done", progress)

            for ebook in group:
                if ebook is None:
                    break
                Logger.ebook = last = ebook

                if not DBUtils.ebook_exists(ebook):
                    continue

                maker = Maker(ebook)

                try:
                    job_queue += maker.mk_job_queue()
                except Exception as what:
                    # report errors, but keep going
                    exception(what)
                    if options.stop_on_errors:
                        return 1

                done_books.append(ebook)

            # send up group
            if options.dry_run:
                info("Job list for #%d - #%d (%d jobs)" % (group[0], last, len(job_queue)))
                print('*' * 80)
                for job in job_queue:
                    print(job)
                    print('*' * 80)
            else:
                info("Calling ebookmaker for #%d - #%d (%d jobs)" %
                     (group[0], last, len(job_queue)))
                setproctitle.setproctitle(
                    "Converting Project Gutenberg #%d - #%d (%d%%)" %
                    (group[0], last, progress)
                )
                run_job_queue(job_queue)

        Notifier.send_notifications(done_books if options.notify else [])

    except KeyboardInterrupt as what:
        # also triggered by: kill -INT(or: kill -2)
        error("User interrupt")
        return 1

    finally:
        os.remove(options.pidfile)

    setproctitle.setproctitle("Cleaning Up")
    info("Program end")
    logging.shutdown()

    return 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

CoverpageWriter.py

Copyright 2022 by Eric Hellman and Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.

Write metadata from source files to the database

"""
from lxml import etree

from libgutenberg.DublinCore import GutenbergDublinCore
from libgutenberg.Logger import info, exception

from ebookmaker import writers


class Writer(writers.BaseWriter):
    """ Class that checks the main file for a book if the db lacks a record. """


    def build(self, job):
        """ checking files for credit metadata """

        try:
            info(etree.tostring(job.dc.to_html()))
            if not bool(job.dc.credit):
                info("Checking credit metadata in file for %s" % job.ebook)
                file_dc = GutenbergDublinCore()
                file_dc.load_from_parser(job.spider.parsers[0])
                info(etree.tostring(file_dc.to_html()))
                if file_dc.credit:
                    job.dc.credit = file_dc.credit
                    job.dc.save(updatemode=1)
                    info("set credit for  %s", job.ebook)
                    return
                info("no credit metadata in %s", job.url)

        except Exception as what:
            exception ("Error checking metadata: %s" % what)
            raise



#!/usr/bin/env python3
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""
UpdateFromWorkflow.py

Copyright 2022 by Eric Hellman and Project Gutenberg
Distributable under the GNU General Public License Version 3 or newer.
scans the backup folder and adds possibly unloaded data.
"""

import os
import json
import re
import stat

from libgutenberg import Logger
from libgutenberg.DublinCore import handle_dc_languages
from libgutenberg.DublinCoreMapping import DublinCoreObject
from libgutenberg.GutenbergGlobals import Struct

from ebookconverter.FileInfo import get_workflow_file, WORKFLOW_LOG_DIR, archive_workflow_file

WORKFLOW_BACKUP_DIR = os.path.join(WORKFLOW_LOG_DIR, 'backup')


def update_from_workflow(dc, record):
    # check if there's a pubinfo, if so, don't update it.
    has_pubinfo = dc.pubinfo and bool(
        dc.pubinfo.publisher or dc.pubinfo.country or dc.pubinfo.years)
    changed = False
    if not has_pubinfo and (not dc.languages or dc.languages[0].id == 'en'):
        set_lang = record.get('LANGUAGE', None)
        if set_lang:
            handle_dc_languages(dc, set_lang)
            changed = True
    if not has_pubinfo:
        dc.pubinfo.publisher = record.get('PUBLISHER', None)
        changed = changed or bool(dc.pubinfo.publisher)
        dc.pubinfo.country = record.get('PUBLISHER_COUNTRY', None)
        changed = changed or bool(dc.pubinfo.country)
        value = record.get('SOURCE_PUBLICATION_YEARS', None)
        if value:
            value = [value] if isinstance(value, str) else value
            if isinstance(value, list):
                for event_year in value:
                    if ':' in event_year:
                        [event, year] = event_year.split(':')
                        dc.pubinfo.years.append((event, year))
                    elif event_year:
                        dc.pubinfo.years.append(('copyright', year))
                changed = True
    if not dc.scan_urls:
        value = record.get('SCANS_ARCHIVE_URL', None)
        if value:
            if isinstance(value, str):
                value = [value]
            if isinstance(value, list):
                for scan_url in value:
                    dc.scan_urls.add(scan_url)
                changed = True
    if not dc.request_key:
        dc.request_key = record.get('REQUEST_KEY', None)
        changed = changed or bool(dc.request_key)
    if not dc.credit:
        dc.credit = record.get('CREDIT', None)
        changed = changed or bool(dc.credit)
    if changed:
        dc.save(updatemode=1)
        Logger.info('updated #%s', dc.project_gutenberg_id)


def main():
    Logger.setup(Logger.LOGFORMAT, 'fileinfo.log')
    Logger.set_log_level(2)

    for filename in sorted(os.listdir(WORKFLOW_LOG_DIR)):
        workflow_file = os.path.join(WORKFLOW_LOG_DIR, filename)
        mode = os.stat(workflow_file)[stat.ST_MODE]
        if stat.S_ISDIR(mode):
            continue
        ebook_num = 0
        m = re.match(r'^(\d+)\.json$', filename)
        if m:
            ebook_num = int(m.group(1))
            Logger.ebook = ebook_num
        else:
            continue

        with open(workflow_file, 'r') as fp:
            wf_data = fp.read()
        data = json.loads(wf_data)
        record = data['DATA']

        dc = DublinCoreObject()
        dc.load_from_database(ebook_num)
        if dc.book:
            update_from_workflow(dc, record)
        archive_workflow_file(workflow_file)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""
AutoRebuild.py

Copyright 2025 by Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.

Looks for any ebooks with significant changed metadata in the last X hours and then
rebuilds the book.
"""

import os
import re
import subprocess

from libgutenberg.Logger import debug, info
from libgutenberg import Logger
from libgutenberg import GutenbergDatabase

from ebookmaker.CommonCode import Options

from ebookconverter.EbookConverter import config

PRIVATE = os.getenv('PRIVATE') or ''
PUBLIC = os.getenv('PUBLIC') or ''

DIRS = PUBLIC + '/dirs'

options = Options()

RE_AUTHOR_BOOK = re.compile(r'fk_books = (\d\d+)')
RE_TITLE_ATTRIBS = re.compile(r'where pk = (\d\d+)')

def get_book_for_attrib(attrib):
    c = GutenbergDatabase.DB.get_cursor()

    c.execute("select fk_books from attributes where pk = %(attrib)s",
               {'attrib': attrib})

    for row in c.fetchall():
        row = GutenbergDatabase.xl(c, row)
        return row.fk_books


def check_sql(sql):
    """ return an ebook number needing rebuild """
    if sql.startswith('update mn_books_authors '):
        # change in author
        match = RE_AUTHOR_BOOK.search(sql)
        if match:
            return match.group(1)
    elif sql.startswith('update attributes set  "fk_attriblist" = 245'):
        # change in title
        match = RE_TITLE_ATTRIBS.search(sql)
        attrib = match.group(1)
        return get_book_for_attrib(attrib)
    return None
        
        


def main():
    goback = 1
    try:
        config()
    except configparser.Error as what:
        error("Error in configuration file: %s", str(what))
        return 1

    Logger.setup(Logger.LOGFORMAT, 'autorebuild.log')
    Logger.set_log_level(2)

    debug("Starting AutoRebuild.py")

    GutenbergDatabase.DB = GutenbergDatabase.Database()
    GutenbergDatabase.DB.connect()
    c  = GutenbergDatabase.DB.get_cursor()

    c.execute("select sql from changelog "
               "where time >= now () - interval '%(goback)s hours' ", {'goback': goback})

    to_rebuild = set()
    for row in c.fetchall():
        row = GutenbergDatabase.xl(c, row)
        to_rebuild.add(check_sql(row.sql))
    range = ''
    for ebook in to_rebuild:
        range = f'{range}{ebook},' if ebook else range
    if range:
        subprocess.call(["ebookconverter", "-v", "--build=all", "--range=%s" % to_rebuild])        
    Logger.ebook = 0
    debug("Done AutoRebuild.py")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: UTF8 -*-

"""

FileInfo.py

Copyright 2009 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Extract metadata from posted PG ebook.

Notes: 10/28/2019

- FileInfo.py looks for files named [number].zip.trig in the dopush "log" directory. If it finds
  one, it looks for files in the in the FILES/[number] directory.
- FileInfo scans the files (or members of the zip file) for a source file containing a plain-text
  "PG header" and puts the metadata from the header into STDOUT.
- When finished, the .trig files are moved to a 'backup' subdirectory.


"""

import json
import os
import re
import shutil
import stat
import sys
import zipfile

import lxml

from libgutenberg.DBUtils import check_session, ebook_exists
from libgutenberg.DublinCoreMapping import DublinCoreObject
from libgutenberg.GutenbergFiles import store_file_in_database, PUBLIC, FILES, FTP
from libgutenberg.GutenbergGlobals import xpath
from libgutenberg.Models import Book
from libgutenberg.Logger import critical, debug, error, exception, info, warning
from libgutenberg import Logger

from ebookmaker.CommonCode import find_candidates

from .Notifier import ADDRESS_BOOK
PRIVATE = os.getenv ('PRIVATE') or ''
PUBLISH = os.getenv ('PUBLISH')  or ''

DOPUSH_LOG_DIR = os.path.join(PRIVATE, 'logs', 'dopush')
WORKFLOW_LOG_DIR = os.path.join(PRIVATE, 'logs', 'json')

PARSEABLE_FILES = '.rst .html .htm .tex .txt'.split ()
HTML_FILES = '.html .htm .xhtml'.split ()



def parseable_file (filename):
    """ Return true if this file should be parsed for a pg header judging by the extension. """
    # these extension are parseable (ascii-ish) files
    return os.path.splitext (filename)[1] in PARSEABLE_FILES


def save_metadata(dc):
    """ Save the metadata. """

    if dc.title:
        dc.title = re.sub (r'\s*\n\s*', '\n', dc.title.strip ())

    if dc.rights.lower ().find ('copyright') > -1:
        dc.rights = 1 # ???

    dc.save()

def get_workflow_file(ebook):
    """ return the workflow metadata file, if it exists"""
    if not ebook:
        return None
    wf_file = os.path.join(WORKFLOW_LOG_DIR, str(ebook) + '.json')
    non_wf_file = os.path.join(WORKFLOW_LOG_DIR, str(ebook) + '.txt')
    debug('wf_file %s exists: %s', wf_file, os.path.exists(wf_file))
    if os.path.exists(wf_file):
        return wf_file
    if os.path.exists(non_wf_file):
        return non_wf_file
    return None

def archive_workflow_file(filename):
    """ archive the workflow metadata file"""
    try:
        shutil.move(filename, os.path.join(WORKFLOW_LOG_DIR, 'backup'))
    except shutil.Error:
        filename_nopath = os.path.split(filename)[1]
        for rev in range(0, 10):
            dest = os.path.join(WORKFLOW_LOG_DIR, 'backup', '%s.%s' % (filename_nopath, rev))
            if not os.path.exists(dest):
                shutil.copy(filename, dest)
                break
            if rev == 9:
                warning('too many revisions, %s not archived', filename)
        os.remove(filename)


def handle_non_dc(data, ebook):
    ww = None
    notify = None
    try:
        non_dc = json.loads(data)
        try:
            notify = non_dc['DATA']['NOTIFY']
        except KeyError:
            info("no notify address for %s", ebook)
        try:
            ww = non_dc['DATA']['WW']
        except KeyError:
            info("no ww address for %s", ebook)

    except ValueError:
        error("bad json file for %s", ebook)

    if notify:
        ADDRESS_BOOK.set_email(ebook, notify)
    if ww:
        ADDRESS_BOOK.set_email(ebook, ww, role='ww')


def scan_header(data, filename, ebook):
    """ Scan pg header in file. Or see if there is a workflow json file. """

    try:
        dc = DublinCoreObject()
        dc.load_book(ebook)
        ext = os.path.splitext(filename)[1]

        workflow_file = get_workflow_file(ebook)
        info('workflow file found: %s', workflow_file)

        if workflow_file and workflow_file.endswith('.json'):
            data = ''
            with open(workflow_file, 'r') as fp:
                data = fp.read()
            try:
                dc.load_from_pgheader(data)
            except ValueError:
                critical("%s was not a valid workflow file", workflow_file)
            if dc.project_gutenberg_id:
                dc.encoding = 'utf-8'
                handle_non_dc(data, dc.project_gutenberg_id)
            archive_workflow_file(workflow_file)
            save_metadata(dc)
            return dc.project_gutenberg_id
        if workflow_file and workflow_file.endswith('.txt'):
            # get text from the non-workflow txt file, if it exists
            with open(workflow_file, 'r') as fp:
                ww = fp.read()
                ADDRESS_BOOK.set_email(ebook, ww, role='ww')
            archive_workflow_file(workflow_file)

        if ext in HTML_FILES:
            body = None
            if data.startswith('<?xml'):
                try:
                    # use XML parser
                    html = lxml.etree.fromstring(data, lxml.etree.XMLParser(load_dtd = True))
                    if html is not None:
                        body = xpath(html, "//xhtml:body")[0]
                        for p in xpath(body, "//xhtml:p"): # fix PGTEI
                            p.tail = "\n"
                except (lxml.etree.ParseError, IndexError) as what:
                    debug("# lxml XMLParser: %s" % what)

            else:
                try:
                    # use HTML parser
                    html = lxml.etree.fromstring(data, lxml.etree.HTMLParser())
                    if html is not None:
                        body = xpath(html, "//body")[0]
                except (lxml.etree.ParseError, IndexError) as what:
                    debug("# lxml HTMLParser: %s" % what)

            if body is not None:
                try:
                    s = lxml.etree.tostring(
                        body,
                        encoding='unicode',
                        method='text')
                    dc.load_from_pgheader(s)
                except UnicodeError:
                    return None

        elif ext == '.rst':
            dc.load_from_rstheader(data)

        else:
            dc.load_from_pgheader(data)

        if str(dc.project_gutenberg_id) == str(ebook):
            save_metadata(dc)
            return dc.project_gutenberg_id
        if dc.project_gutenberg_id:
            error('loaded gutenberg id %s in %s did not match trigger id %s',
                  dc.project_gutenberg_id, filename, ebook)
        return None

    except ValueError as what:
        exception (what)
        debug("Could not scan header of %s" % filename)
        return None


def scan_file(filename, ebook):
    """ Scan one file. """
    data = ''
    if parseable_file(filename):
        with open(filename, 'rb') as fp:
            data = read_string(fp.read())
    if scan_header(data, filename, ebook):
        return True
    return False


def scan_zip(filename, ebook):
    """ Scan a zip file like a dir. """

    try:
        zip_ = zipfile.ZipFile(filename)
        for member in sorted(zip_.namelist(), key=file_sort_key):
            if parseable_file(member):
                if scan_header(read_string(zip_.read(member)), member, ebook):
                    return True
    except (zipfile.error, NotImplementedError):
        warning('zipfile opening error for %s', filename)
    return False

def read_string(bytes_data):
    for encoding in ['utf-8', 'iso-8859-1']:
        try:
            data = bytes_data.decode(encoding)
            return data
        except UnicodeError:
            pass
    return ''

def is_readable(filename):
    """ Used to be "stat_file. The 'stat' part has been refactored into libgutenberg """
    return os.access(filename, os.R_OK)


def file_sort_key(filename):
    """ Sort files according to metadata quality.

    Files with best metadata first. """

    name, ext = os.path.splitext(filename)

    # sort parseable files first
    if ext in PARSEABLE_FILES:
        # -z sorts ascii text files last
        return "%02d %s-z" % (PARSEABLE_FILES.index(ext), name)
    return '99' + name + '-z'


def create_ebook(ebook_num, session=None):
    """ create an ebook, only info will be the number
    release_date=today() autoassigned by server"""
    session = check_session(session)
    new_ebook = Book(pk=ebook_num)
    session.add(new_ebook)
    session.commit()


def scan_directory(ebook_num):
    """ Scan one directory in the new archive filesystem. """
    # is ebook_num in db?

    dirname = os.path.join(FILES, str(ebook_num))
    debug("Scanning directory %s ...", dirname)

    found_files = find_candidates(dirname, file_filter=is_readable)

    header_found = False
    for filename in sorted(found_files, key=file_sort_key):
        if not ebook_exists(ebook_num):
            create_ebook(ebook_num)
        if not header_found:
            if filename.endswith('.zip'):
                header_found = scan_zip(filename, ebook_num)
            else:
                header_found = scan_file(filename, ebook_num)

        store_file_in_database(ebook_num, filename, None)

def scan_dopush_log():
    """ Scan the dopush log directory for new files.

    Files in this directory are placeholders only. The real files are
    in FILES  and PUBLISH/.

    """

    retcode = 1

    for filename in sorted(os.listdir(DOPUSH_LOG_DIR)):
        mode = os.stat(os.path.join(DOPUSH_LOG_DIR, filename))[stat.ST_MODE]
        if stat.S_ISDIR(mode):
            continue

        debug("Tag file: %s" % filename)

        ebook_num = 0
        m = re.match(r'^(\d+)\.zip\.trig$', filename)
        if m:
            ebook_num = int(m.group(1))
            Logger.ebook = ebook_num
            scan_directory(ebook_num)

        shutil.move(os.path.join(DOPUSH_LOG_DIR, filename),
                     os.path.join(DOPUSH_LOG_DIR, 'backup', filename))
        retcode = 0

    return retcode

def main():
    Logger.setup(Logger.LOGFORMAT, 'fileinfo.log')
    Logger.set_log_level(2)

    # This is the only encoding used at PG/ibiblio.
    # Bail out early if the environment is misconfigured.
    assert str(sys.stdout.encoding).lower() == 'utf-8'

    if len(sys.argv) > 1:
        for arg in sys.argv[1:]:
            try:
                Logger.ebook = int(arg)
                scan_directory(int(arg))
            except ValueError: # no int
                scan_file(arg, None)

    else:
        sys.exit(scan_dopush_log())

if __name__ == '__main__':
    main()



#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
BlueskyWriter.py

Copyright 2025 by Project Gutenberg

Distributable under the GNU General Public License Version 3 or newer.

Skeet about new ebooks.

"""

import os
from datetime import datetime, timezone

import requests

from requests import RequestException

from libgutenberg.Logger import error, exception
from libgutenberg.GutenbergGlobals import SkipOutputFormat
from libgutenberg import GutenbergDatabase

from ebookmaker import writers
from ..EbookConverter import make_output_filename


URL = 'https://bsky.social/xrpc/com.atproto.server.createSession'
SKEETURL = "https://bsky.social/xrpc/com.atproto.repo.createRecord"
BLOBURL = "https://bsky.social/xrpc/com.atproto.repo.uploadBlob"

HANDLE = 'new.gutenberg.org'
APP_PASSWORD = os.environ.get('BLUESKY_APP_PASSWORD')


class Writer(writers.BaseWriter):
    """ Class to skeet about new ebooks.

    """

    __instance = None
    __session = None

    def __new__(cls):
        """ Make Writer a singleton. """

        if cls.__instance is None:
            cls.__instance = object.__new__(cls)
        return cls.__instance


    def __init__(self):

        super(Writer, self).__init__()

        
        if not APP_PASSWORD:
            raise SkipOutputFormat ('Bluesky credentials not found.')
            
    def get_session(cls):
        if cls.__session:
            return cls.__session
        
        resp = requests.post(URL, json={'identifier': HANDLE, 'password': APP_PASSWORD})
        resp.raise_for_status()
        cls.__session = resp.json()
        return cls.__session

    def get_blob(self, img_path):
        """ get cover image, post it to get a blob identifier """

        session = self.get_session()
        with open(img_path, "rb") as f:
            img_bytes = f.read()
        
        # this size limit is specified in the app.bsky.embed.images lexicon
        if len(img_bytes) > 1000000:
            raise Exception(
                f"image file size too large. 1000000 bytes maximum, got: {len(img_bytes)}"
            )
        
        resp = requests.post(BLOBURL,
            headers={
                "Content-Type": "image/jpeg",
                "Authorization": "Bearer " + session["accessJwt"],
            },
            data=img_bytes,
        )
        resp.raise_for_status()
        return resp.json()["blob"]

    def build(self, job):
        """ Skeet. """
        
        pg_id = job.dc.project_gutenberg_id
        coverfile = make_output_filename('cover.medium', pg_id)
        coverpath = os.path.join(job.outputdir, coverfile)

        if pg_id < 75000:
            # don't skeet about old books
            return

        db = GutenbergDatabase.Database()
        db.connect()
        c = db.get_cursor()

        c.execute("select * from tweets where media = 'Bluesky' and fk_books = %(fk_books)s",
                   {'fk_books': pg_id})

        rows = c.fetchone()
        if rows:
            # already skeeted
            return

        try:
            session = self.get_session()
            try:
                blob = self.get_blob(coverpath)
            except (FileNotFoundError) as what:
                exception('BlueskyWriter: could not find cover file (#%s)' % what)
                blob = ""

            title = job.dc.make_pretty_title(300 - 84).replace('"', '\\"')  # 300 is max length
            headers = {'Authorization': 'Bearer ' + session["accessJwt"]}
            pg_url = f'https://www.gutenberg.org/ebooks/{pg_id}'
            skeet = f"New #ebook at Project Gutenberg: {title} {pg_url}"
            skeetlen = len(skeet.encode("UTF-8"))
            facets =  [{
                "index": {
                    "byteStart": skeetlen - len(pg_url),
                    "byteEnd": skeetlen
                },
                "features": [{
                  "$type": "app.bsky.richtext.facet#link",
                  "uri": pg_url
                }]
            }]
            embed = {"$type": "app.bsky.embed.images",
                     "images": [{"alt": "", "image": blob,}],
                    }
            lang_id = job.dc.languages[0].id if len(job.dc.languages) else 'en'
            post = {"$type": "app.bsky.feed.post",
                    "text": skeet,
                    "createdAt": datetime.now(timezone.utc).isoformat().replace("+00:00", "Z"),
                    "langs": [lang_id],
                    "facets": facets,
                    }
            if blob:
                post['embed'] = embed
            data = {"repo": session["did"],
                    "collection": "app.bsky.feed.post",
                    "record": post,
                    }

            r = requests.post(SKEETURL, headers=headers, json=data)
            r.raise_for_status()

            c.execute('start transaction')
            c.execute(
                """\
                insert into tweets (fk_books, media, time)
                values (%(fk_books)s, 'Bluesky', now ())
                """, {'fk_books': job.dc.project_gutenberg_id})
            c.execute('commit')

        except (RequestException) as what:
            exception('BlueskyWriter: could not skeet about ebook: %s' % what)

        except (GutenbergDatabase.DatabaseError) as what:
            c.execute('rollback')
            exception('BlueskyWriter: could not write to database (#%s)' % what)




#!/usr/bin/env python
#  -*- mode: python; indent-tabs-mode: nil; -*- coding: utf-8 -*-

"""
FacebookWriter.py

Copyright 2010 by Marcello Perathoner

Distributable under the GNU General Public License Version 3 or newer.

Cackle about new ebooks.
"""


import os

import requests

from libgutenberg import GutenbergDatabase
from libgutenberg.Logger import exception, error, info, debug
from libgutenberg.GutenbergGlobals import SkipOutputFormat

from ebookmaker import writers

PGURL = "https://www.gutenberg.org/"

GRAPH = "https://graph.facebook.com/v19.0/"

class Writer (writers.BaseWriter):
    """ Class to post about new ebooks.

    """

    __instance = None

    def __new__ (cls):
        """ Make Writer a singleton. """

        if cls.__instance is None:
            cls.__instance = object.__new__ (cls)
        return cls.__instance


    def __init__ (self):
        super (Writer, self).__init__ ()

        # default = New Project Gutenberg Books Page
        self.page_id = os.environ.get ('FB_PAGE_ID','171646249561837') 
        self.access_token

    _access_token = None
    @property
    def access_token (self):
        if not self._access_token:
            token_file = os.path.expanduser ('~/.fb_access_token')
            with open (token_file, 'r') as tf:
                self._access_token = tf.read ().strip ()
            if not self._access_token:
                raise SkipOutputFormat ('Facebook credentials not found.')
        return self._access_token


    def build (self, job):
        """ Post. """

        if job.dc.project_gutenberg_id < 45000:
            # don't post about old books
            info ('old book, no Facebook post')
            return

        db = GutenbergDatabase.Database ()
        db.connect ()
        c = db.get_cursor ()

        c.execute ("select * from tweets where media = 'Facebook' and fk_books = %(fk_books)s",
                   { 'fk_books': job.dc.project_gutenberg_id })

        rows = c.fetchone ()
        if rows:
            # already posted
            info ('already posted, no new Facebook post')
            return

        try:
            id_ = job.dc.project_gutenberg_id
            data = {
                'link': PGURL + "ebooks/%d" % id_,
                'message': "This is a free ebook by Project Gutenberg.",
                'access_token': self.access_token,
                }

            fp = requests.post (
                "%s%s/feed" % (GRAPH, self.page_id),
                data = data,
                timeout = 60
            )
            if fp.status_code == 200:
                c.execute ('start transaction')
                c.execute ("insert into tweets (fk_books, media, time) values (%(fk_books)s, 'Facebook', now ())",
                           { 'fk_books': job.dc.project_gutenberg_id })
                c.execute ('commit')
                info ("FacebookWriter: posted ebook: %d" % id_)
            else:
                error ("Facebook returned HTTP code %d (%s)" % (fp.status_code, fp.text))

        except GutenbergDatabase.DatabaseError as what:
            c.execute ('rollback')
            exception ('FacebookWriter: could not post ebook: %s' % (what))
